#!/usr/bin/env python
#
# Data Quality Checker also known as DQC
#
# Read the README.md for more detailed information
#
#       Barcelona Supercomputing Center - Centro Nacional de Supercomputacion
#       Earth Sciences Department - Computational Earth Sciences
#     
#     Authors: 
#       Joan Sala Calero (joan.sala@bsc.es)
#       Jesus Pena-Izquierdo (jesus.pena@bsc.es)

import os
import pandas as pd
import numpy as np

from bokeh.plotting import figure
from bokeh.layouts import gridplot, column
from bokeh.io import show, save, output_file
from bokeh.models import ColumnDataSource, HoverTool, Span

from utils.utils_files import *

"""
Class: Bokeh plots
---
- General plot functions for histograms 
- Statistical formulas needed to produce the plots
"""
class bokehPlot:

    def __init__(self, outfname, csvs, csvh, vv, vmin, vmax):
        self.outfile_hist = outfname
        self.outfile_stat = outfname.replace('.html', '_boxplot.html')
        self.csv_hist = csvh
        self.csv_stat = csvs
        self.varname = vv
        self.valid_min = vmin
        self.valid_max = vmax
        self.regions = {
          'total' : '', 
          'N_polar' : '[60 to 90 deg]', 
          'N_midlat' : '[30 to 60 deg]', 
          'N_hadley' : '[0 to 30 deg]', 
          'S_hadley' : '[0 to -30 deg]', 
          'S_midlat' : '[-30 to -60 deg]', 
          'S_polar' : '[-60 to -90 deg]'
        }

    """
    Plot a boxplot diagram
    """
    def plot_boxplot(self, col='File', pw=1000, ph=550):

      # generate some synthetic time series for six different categories
      df = pd.read_csv(self.csv_stat)
      tt = get_dates_from_files(df[col])
      if len(tt) > 2:
        tdiff = np.diff(tt)[0] * 0.5
      else:
        tdiff = 10 

      # Define figure
      pw=1000
      ph=600

      # Set up the figure same as before
      p = figure(plot_width = pw, plot_height = ph,
                 title = 'Observed values boxplot',
                 x_axis_label = 'Time', 
                 y_axis_label = self.varname,
                 x_axis_type='datetime')

      for ii, r in df.iterrows():

          upper = r['q3']+1.5*r['IQD']
          lower = r['q1']-1.5*r['IQD']

          # stems
          p.segment(tt[ii], upper, tt[ii], r['q3'], line_color="black", legend_label='-/+ 1.5*IQD')
          p.segment(tt[ii], lower, tt[ii], r['q1'], line_color="black", legend_label='-/+ 1.5*IQD')

          # boxes
          p.vbar(x=tt[ii], bottom=r['Med'], top=r['q3'], width=tdiff, fill_color="#E08E79", line_color="black", legend_label='Q3 (75%)')
          p.vbar(x=tt[ii], bottom=r['q1'], top=r['Med'], width=tdiff, fill_color="#3B8686", line_color="black", legend_label='Q1 (25%)')

          # whiskers (almost-0 height rects simpler than segments)
          p.rect(tt[ii], lower, tdiff*0.5, 0.01, line_color="black")
          p.rect(tt[ii], upper, tdiff*0.5, 0.01, line_color="black")

      # Min/Max lines
      p.line(tt, df['Min'], color='green', legend_label='Observed min')
      p.line(tt, df['Max'], color='red', legend_label='Observed max')

      # Write plot
      print('INFO: Saving plot {}'.format(self.outfile_stat))
      output_file(self.outfile_stat)
      save(p)

    """
    Add valid min/max to plot (vertical lines + highlights)
    """
    def add_valid_minmax(self, fig, l, y, nz):

        vert_lines = []

        if not(self.valid_max is None):
          vert_lines.append(Span(location=self.valid_max, dimension='height', line_color='red', line_dash='dashed', line_width=1.5))       
          vmax = np.where(l[nz] > self.valid_max)
          if len(vmax):
            fig.circle((l[nz])[vmax], (y[nz])[vmax], color='red', fill_color="white", legend_label='outliers')

        if not(self.valid_min is None):
          vert_lines.append(Span(location=self.valid_min, dimension='height', line_color='red', line_dash='dashed', line_width=1.5))        
          vmins = np.where(l[nz] < self.valid_min)
          if len(vmins):
            fig.circle((l[nz])[vmins], (y[nz])[vmins], color='red', fill_color="white", legend_label='outliers')
        
        # Add vertical lines
        if vert_lines != []:
          fig.renderers.extend(vert_lines)

    """
    Plot a single histogram, return figure
    """
    def plot_histogram_figure(self, df, tag='total', col='File', pw=500, ph=250):

        # Get bins (left and right) + remove first element
        l = np.delete(df.columns.values, 0).astype(float) 
        xdist = abs(l[0]-l[1])
        r = [ l+xdist for v in l ]
        
        # Get count + remove first element
        y = np.delete(df.loc[df[col] == tag].values, 0).astype(float)

        # Get non-zero bins
        nz = np.nonzero(y)

        # Set up the figure same as before
        fig = figure(plot_width = pw, plot_height = ph,
                   title = 'Observed values histogram {}'.format(self.regions[tag]),
                   x_axis_label = self.varname, 
                   y_axis_label = 'Count')

        # Histograms and points where we have values
        fig.vbar(x=l[nz], width=xdist, bottom=0, top=y[nz], color="#941333")
        fig.dash(x=l[nz], y=y[nz], color="#941333")
        fig.toolbar.logo = None
        
        # Vertical lines indicating valid_min/valid_max if available
        self.add_valid_minmax(fig, l, y, nz)
        fig.legend.visible = (tag=='total')

        return fig

    """
    Plot multiple histograms into a grid and return a grid of the plots (file) 
    """
    def plot_histograms(self):

        # Read histograms into pandas
        df_hist = pd.read_csv(self.csv_hist)

        # Total histogram
        p0 = self.plot_histogram_figure(df_hist, tag='total', pw=1000, ph=450)

        # Latitude regions
        p1 = None
        try:
          figs = []
          for t in [ 'N_polar', 'S_polar', 'N_midlat', 'S_midlat', 'N_hadley', 'S_hadley' ]:
            figs.append(self.plot_histogram_figure(df_hist, tag=t))

          # Group plots into grid
          p1 = gridplot(figs, ncols=2, toolbar_options=dict(logo=None))
        except:
          pass # Dataset has no regions

        # Save to file        
        print('INFO: Saving plot {}'.format(self.outfile_hist))
        output_file(self.outfile_hist)
        
        if p1 is None:
          save(p0)
        else:
          save(column(p0, p1))
